function [g,d,info] = tvnorm(G,Tspan,varargin)
% XXX Add Documentation
%   [g,d,info] = tvnorm(G,Tspan,NormType,Opt)
% Type = 'L2toL2' or 'L2toE'. This input is optional with default 'L2toL2'.
% Opt is currently the ODE options.
%
% XXX Add inputs to specify tolerance, display, and bisection bounds?
%   The easiest way would be to make Opt a structure (or object) with
%   fields Opt.ODEOpt, Opt.Tol, Opt.Bounds, Opt.Display


%% Input Processing
narginchk(2,4);
nin = nargin;
nout = nargout;

vin = varargin;
NormType = [];
Opt = [];
if nin==3
    if isa(vin{end},'struct')
        Opt = vin{1};
    else
        NormType = vin{1};
    end
else
    [NormType,Opt] = deal(vin{:});
end
if isempty(NormType)
    NormType = 'L2toL2';
end
if isscalar(Tspan)
    T0=0; 
    Tf=Tspan;
else
    T0 = Tspan(1);
    Tf = Tspan(2);
end

%% Build Cost Function Matrices
%   R(t,g) = R0(t) - g^2 I
[A,B,C,D] = ssdata(G);
[Nx,Nu] = size(B);
E = [];
if isequal(NormType,'L2toL2')
    Q = C'*C;
    S = C'*D;
    R0 = D'*D;
    F = zeros(Nx);
else
    % XXX Verify D(t)==0 for L2 to E    
    % XXX This set-up requires R(t) to be inverted at each time step.
    % However, no inversion is actually required since we can set
    % R=I and B'=B/gamma.  This should speed up the integration.
    Q = zeros(Nx);
    S = [];
    R0 = zeros(Nu);    
    CTf = evalt(C,Tf);
    F = CTf'*CTf;
end

%% Bisection Initialization
dflag = true;
gUpp = [];
tol = 1e-3;

%% Lower Bound Phase:
%  Require R(t,g)=R0(t)-g^2*I<0 for all t in [0,T]
% XXX This does not work on Function data types. Also, it assumes
% the time grid (for 'Grid' data types) is sufficiently fine.
if isequal(NormType,'L2toL2')
    ev = eig(R0);
    evmax = max( ev.Data(:) );  % XXX Implement a TVMAX function?
    gLow = sqrt(evmax);
else
    gLow = 0;
end
if dflag
    fprintf('\n Lower Bound = %4.3f',gLow);
end

%% Upper Bound Phase
if isempty(gUpp)
    gUpp = 2*gLow+1;  % XXX Better choice?
end
cnt = 0;
cntMax = 20;
haveUpper = false;
PUpp = []; PdotUpp = []; solUpp = [];
PLow = []; PdotLow = []; solLow = [];
while ~haveUpper && cnt<=cntMax
    % Pick a gamma
    gTry = gUpp;
    
    % Solve LTV Riccati Equation
    R = R0-gTry^2*eye(Nu);
    if nout==1
        P = crde(A,B,Q,R,S,E,F,[Tf T0],Opt);
        Pdot = []; sol = [];
    else
        [P,~,Pdot,sol] = crde(A,B,Q,R,S,E,F,[Tf T0],Opt);
    end
    
    % Check convergence of P
    if P.Time(1)>T0
        % P did not converge
        gUpp = 2*gUpp;
        gLow = gTry; 
        PLow = P; PdotLow = Pdot; solLow = sol;
    else
        % P converged
        haveUpper = true;
        gUpp = gTry; 
        PUpp = P; PdotUpp = Pdot; solUpp = sol;
    end
    cnt = cnt+1;
end

if ~haveUpper
    % Could not find a finite upper bound
    if dflag
        fprintf(['\n Could not find a finite upper bound.' ...
            ' Infeasible at gTry = %4.1f\n'],gUpp);
    end
    g = inf; d=[]; info=[];
    return
end
if dflag
    fprintf('\n Upper Bound = %4.3f',gUpp);
end


%% Bisection Phase
if dflag
    fprintf('\n');
end
while (gUpp - gLow > tol*gUpp) && gUpp>1e-4
    gTry = (gUpp+gLow)/2;
    
    % Solve LTV Riccati Equation
    R = R0-gTry^2*eye(Nu);
    if nout==1
        P = crde(A,B,Q,R,S,E,F,[Tf T0],Opt);
        Pdot = []; sol = [];
    else
        [P,~,Pdot,sol] = crde(A,B,Q,R,S,E,F,[Tf T0],Opt);
    end
    
    % Check convergence of P
    if P.Time(1)>T0
        % P did not converge
        gLow = gTry;
        PLow = P; PdotLow = Pdot; solLow = sol;
        if dflag
            fprintf(' gTry = %4.3f \t Conv=N \t T0= %4.3f \t gLow = %4.3f \t gUpp = %4.3f\n', ...
                gTry,P.Time(1),gLow,gUpp);
        end
    else
        % P converged
        gUpp = gTry;
        PUpp = P; PdotUpp = Pdot; solUpp = sol;
        if dflag
            fprintf(' gTry = %4.3f \t Conv=Y \t T0= %4.3f \t gLow = %4.3f \t gUpp = %4.3f\n', ...
                gTry,P.Time(1),gLow,gUpp);
        end
    end
    
end

%% Store Final Result
if isempty(gUpp)
    % No Upper Bound Found
    g = inf;    
else
    g = gUpp;    
end

info.Lower.Gain = gLow;
info.Lower.P = PLow;
info.Lower.Pdot = PdotLow;
info.Lower.sol = solLow;

info.Upper.Gain = g;
info.Upper.P = PUpp;
info.Upper.Pdot = PdotUpp;
info.Upper.sol = solUpp;


%% Construct Worst-Case Input
% The construction is based on the two-point boundary value problem 
% (TPBVP) related to the (Q,S,R,F) cost.  This particular implementation
% using the transformed Hamiltonian dynamics was developed by A. Iannelli 
% and P. Seiler in September 2017.
d = [];
if nout>=2 && ~isempty(PLow)
    % Compute largest eigenvalue of PLow     
    T = PLow.Time;
    [evecP,evalP] = eig( evalt(PLow,T(1)) );
    [emax,idx] = max( diag(evalP) );
    vmax = evecP(:,idx);

    % Evaluate State/Cost Matrices on Same Time Grid as P
    R = tvmat( R0-gLow^2*eye(Nu) );
    A = evalt(A,T,'tvmat');
    B = evalt(B,T,'tvmat');    
    if isequal(NormType,'L2toL2')
        R = evalt(R,T,'tvmat');
        S = evalt(tvmat(S),T,'tvmat');  
    else
        S = tvmat(zeros(Nx,Nu));
    end
    
    % Simulate Transformed Hamiltonian system 
    M = inv(R)*(PLow*B+S)';
    H11 = A-B*M;   
    odefh = @(t,E) evalt(H11,t)*E;
    E0 = vmax/emax;
    [tE,E] = ode45( odefh,T,E0);
    E = tvmat(E,tE);
        
    % Construct disturbance
    % Note: This calculation uses a non-convergent Riccati solution and
    % hence the disturbance starts at PLow.Time(1) > T0.
    d = -M*E;
    
    % Append zero input to disturbance on the window [PLow.Time(1) T0)
    dTime = d.Time;
    dData = d.Data;

    Nd = size(B,2);
    dTime = [T0; T0+0.999*(dTime(1)-T0); dTime];
    dData = cat(3,zeros(Nd,1,2), dData);
    d = tvmat(dData,dTime);
end